; Note: The license below is based on the template at:
; http://opensource.org/licenses/BSD-3-Clause

; Copyright (C) 2023 Regents of the University of Texas
; All rights reserved.

; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:

; o Redistributions of source code must retain the above copyright
;   notice, this list of conditions and the following disclaimer.

; o Redistributions in binary form must reproduce the above copyright
;   notice, this list of conditions and the following disclaimer in the
;   documentation and/or other materials provided with the distribution.

; o Neither the name of the copyright holders nor the names of its
;   contributors may be used to endorse or promote products derived
;   from this software without specific prior written permission.

; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
; HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

; Original Author(s):
; Mertcan Temel         <mert@utexas.edu>

(in-package "RP")

(include-book "projects/rp-rewriter/lib/mult3/top" :dir :system)

;; Increase rewriter step limit. Usually only necessary for very large multipliers. 
(rp::set-rw-step-limit (expt 2 30))
(rp::set-rp-backchain-limit 10000)

;; increase the maximum memory to 16GB that triggers garbage collection (only for CCL) 
(value-triple (acl2::set-max-mem (* 16 (expt 2 30))))


;; ---------------------------------------------------------------------------

;; Example 1
(verify-svtv-of-mult :name DT-SB4-BK-64x64
                     :concl (equal design_res
                                   (loghead 128 (* (logext 64 in1)
                                                   (logext 64 in2))))
                     :read-from-file "parsed/")

;; logext means sign-extend, loghead means zero-extend. 

;; ---------------------------------------------------------------------------


;; Example 2
;; When trying to verify the buggy design only with vescmul, it will fail
;; quickly and return the rewritten term. After you submit the event, if you
;; scroll up, you'll see the simplified term and the error message. 
(verify-svtv-of-mult :name buggy-DT-SB4-BK-64x64
                      :concl (equal design_res
                                    (loghead 128 (* (logext 64 in1)
                                                    (logext 64 in2))))
                      :read-from-file "parsed/"

                      ;; setting keep-going will prevent this event from
                      ;; halting certification of this file. 
                      :keep-going t)


;; Before we proceed  to generate counterexample, let's configure  the SAT call
;; and ask the program to run "glucose"  on the shell for glucose SAT solver. You
;; may change this to a different sat solver.
(local
 (progn
   (defun my-sat-config ()
     (declare (xargs :guard t))
     (satlink::make-config :cmdline "glucose "
                           :verbose t
                           :mintime 1/2
                           :remove-temps t))
   (defattach fgl::fgl-satlink-config my-sat-config)))

;; now let's set ":then-fgl t" and this time get a counterexample. 
(verify-svtv-of-mult :name buggy-DT-SB4-BK-64x64
                     :then-fgl t
                     :concl (equal design_res
                                   (loghead 128 (* (logext 64 in1)
                                                   (logext 64 in2))))
                     :read-from-file "parsed/"

                     :keep-going t)


;; ---------------------------------------------------------------------------

;; Example 3
;; multiply-add can be verified as well.
(verify-svtv-of-mult :name MAC_DT_SB8_HC_32x32_plus_32_63to0
                     :concl (equal design_res
                                   (loghead 64 (+ (* (logext 32 in1)
                                                     (logext 32 in2))
                                                  (logext 32 in3))))
                     :read-from-file "parsed/")


;; ---------------------------------------------------------------------------

;; Example 4

;; Arbitrary configurations are also supported

(verify-svtv-of-mult :name DOT_Product_WT_SB4_JSkCond_6_11x9_plus_23_22to3
                     :concl (and
                             ;; openly state the spec. ash is arithmetic shift. 
                             (equal design_res
                                    (ash
                                     (loghead 23
                                              (+ (* (logext 11 in1) (logext 9 in2))
                                                 (* (logext 11 (ash in1 -11)) (logext 9 (ash in2 -9)))
                                                 (* (logext 11 (ash in1 -22)) (logext 9 (ash in2 -18)))
                                                 (* (logext 11 (ash in1 -33)) (logext 9 (ash in2 -27)))
                                                 (* (logext 11 (ash in1 -44)) (logext 9 (ash in2 -36)))
                                                 (* (logext 11 (ash in1 -55)) (logext 9 (ash in2 -45)))
                                                 (logext 23 in3)))
                                     -3))
                             ;; OR the design also provides a
                             ;; high-level specification. We can check
                             ;; that. 
                             (equal design_res  spec_res))
                     :read-from-file "parsed/")



